# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
import warnings

import matplotlib.pyplot as plt
import numpy as np

from pulser.utils import validate_duration


class Waveform(ABC):
    """The abstract class for a pulse's waveform."""

    def __init__(self, duration):
        """Initializes a waveform with a given duration.

        Args:
            duration (int): The waveforms duration (in multiples of 4 ns).
        """
        self._duration = validate_duration(duration, min_duration=16,
                                           max_duration=4194304)

    @property
    @abstractmethod
    def duration(self):
        """The duration of the pulse (in multiples of 4 ns)."""
        pass

    @property
    @abstractmethod
    def samples(self):
        """The value at each time step that describes the waveform.

        Returns:
            samples(np.ndarray): A numpy array with a value for each time step.
        """
        pass

    @property
    def first_value(self):
        """The first value in the waveform."""
        return self.samples[0]

    @property
    def last_value(self):
        """The last value in the waveform."""
        return self.samples[-1]

    @property
    def integral(self):
        """Integral of the waveform (time in ns, value in rad/µs)."""
        return np.sum(self.samples) * 1e-3  # ns * rad/µs = 1e-3

    def draw(self):
        """Draws the waveform."""

        fig, ax = plt.subplots()
        self._plot(ax, "rad/µs")

        plt.show()

    @abstractmethod
    def __str__(self):
        pass

    @abstractmethod
    def __repr__(self):
        pass

    def __eq__(self, other):
        if not isinstance(other, Waveform):
            return False
        elif self.duration != other.duration:
            return False
        else:
            return np.all(np.isclose(self.samples, other.samples))

    def __hash__(self):
        return hash(tuple(self.samples))

    def _plot(self, ax, ylabel, color=None):
        ax.set_xlabel('t (ns)')
        ts = np.arange(self.duration)
        if color:
            ax.set_ylabel(ylabel, color=color, fontsize=14)
            ax.plot(ts, self.samples, color=color)
            ax.tick_params(axis='y', labelcolor=color)
            ax.axhline(0, color=color, linestyle=':', linewidth=0.5)
        else:
            ax.set_ylabel(ylabel, fontsize=14)
            ax.plot(ts, self.samples)
            ax.axhline(0, color='black', linestyle=':', linewidth=0.5)


class CompositeWaveform(Waveform):
    """A waveform combining multiple smaller waveforms.

    Args:
        waveforms(Waveform): Two or more waveforms to combine.
    """

    def __init__(self, *waveforms):
        """Initializes a waveform from multiple waveforms."""
        if len(waveforms) < 2:
            raise ValueError("Needs at least two waveforms to form a "
                             "CompositeWaveform.")
        for wf in waveforms:
            self._validate(wf)

        self._waveforms = list(waveforms)

    @property
    def duration(self):
        """The duration of the pulse (in multiples of 4 ns)."""
        duration = 0
        for wf in self._waveforms:
            duration += wf.duration
        return duration

    @property
    def samples(self):
        """The value at each time step that describes the waveform.

        Returns:
            numpy.ndarray: A numpy array with a value for each time step.
        """
        return np.concatenate([wf.samples for wf in self._waveforms])

    @property
    def first_value(self):
        """The first value in the waveform."""
        return self._waveforms[0].first_value

    @property
    def last_value(self):
        """The last value in the waveform."""
        return self._waveforms[-1].last_value

    @property
    def waveforms(self):
        """The waveforms encapsulated in the composite waveform."""
        return list(self._waveforms)

    def insert(self, waveform, where=0):
        """Insert a new waveform into the CompositeWaveform.

        Args:
            waveform: A valid waveform.

        Keyword Args:
            where (default=0): Index before which the waveform is inserted.
        """
        self._validate(waveform)
        self._waveforms.insert(where, waveform)

    def append(self, waveform):
        """Append a new waveform to the end of a CompositeWaveform.

        Args:
            waveform: A valid waveform.
        """
        self.insert(waveform, where=len(self._waveforms))

    def _validate(self, waveform):
        if not isinstance(waveform, Waveform):
            raise TypeError("{!r} is not a valid waveform. Please provide a "
                            "valid Waveform.".format(waveform))

    def __str__(self):
        contents = ["{!r}"] * len(self._waveforms)
        contents = ", ".join(contents)
        contents = contents.format(*self._waveforms)
        return f'Composite({contents})'

    def __repr__(self):
        return f'CompositeWaveform({self.duration} ns, {self._waveforms!r})'


class CustomWaveform(Waveform):
    """A custom waveform.

    Args:
        samples (array_like): The modulation values at each time step
            (in rad/µs). The number of samples dictates the duration, so its
            length has to be a multiple of 4.
    """

    def __init__(self, samples):
        """Initializes a custom waveform."""
        samples_arr = np.array(samples)
        self._samples = samples_arr
        with warnings.catch_warnings():
            warnings.filterwarnings("error")
            try:
                super().__init__(len(samples))
            except UserWarning:
                raise ValueError("The provided samples correspond to a "
                                 "waveform of invalid duration. Please give"
                                 " samples whose size is a multiple of 4.")

    @property
    def duration(self):
        """The duration of the pulse (in multiples of 4 ns)."""
        return self._duration

    @property
    def samples(self):
        """The value at each time step that describes the waveform.

        Returns:
            numpy.ndarray: A numpy array with a value for each time step.
        """
        return self._samples

    def __str__(self):
        return 'Custom'

    def __repr__(self):
        return f'CustomWaveform({self.duration} ns, {self.samples!r})'


class ConstantWaveform(Waveform):
    """A waveform of constant value.

    Args:
        duration: The waveform duration (in multiples of 4 ns).
        value: The modulation value (in rad/µs).
    """

    def __init__(self, duration, value):
        """Initializes a constant waveform."""
        super().__init__(duration)
        self._value = float(value)

    @property
    def duration(self):
        """The duration of the pulse (in multiples of 4 ns)."""
        return self._duration

    @property
    def samples(self):
        """The value at each time step that describes the waveform.

        Returns:
            numpy.ndarray: A numpy array with a value for each time step.
        """
        return np.full(self.duration, self._value)

    @property
    def first_value(self):
        """The first value in the waveform."""
        return self._value

    @property
    def last_value(self):
        """The last value in the waveform."""
        return self._value

    def __str__(self):
        return f"{self._value:.3g} rad/µs"

    def __repr__(self):
        return (f"ConstantWaveform({self._duration} ns, "
                + f"{self._value:.3g} rad/µs)")


class RampWaveform(Waveform):
    """A linear ramp waveform.

    Args:
        duration: The waveform duration (in multiples of 4 ns).
        start: The initial value (in rad/µs).
        stop: The final value (in rad/µs).
    """

    def __init__(self, duration, start, stop):
        """Initializes a ramp waveform."""
        super().__init__(duration)
        self._start = float(start)
        self._stop = float(stop)

    @property
    def duration(self):
        """The duration of the pulse (in multiples of 4 ns)."""
        return self._duration

    @property
    def samples(self):
        """The value at each time step that describes the waveform.

        Returns:
            numpy.ndarray: A numpy array with a value for each time step.
        """
        return np.linspace(self._start, self._stop, num=self._duration)

    @property
    def slope(self):
        r"""Slope of the ramp, in :math:`s^{-15}`."""
        return (self._stop - self._start) / self._duration

    @property
    def first_value(self):
        """The first value in the waveform."""
        return self._start

    @property
    def last_value(self):
        """The last value in the waveform."""
        return self._stop

    def __str__(self):
        return f"Ramp({self._start:.3g}->{self._stop:.3g} rad/µs)"

    def __repr__(self):
        return (f"RampWaveform({self._duration} ns, " +
                f"{self._start:.3g}->{self._stop:.3g} rad/µs)")


class BlackmanWaveform(Waveform):
    """A Blackman window of a specified duration and area.

    Args:
        duration: The waveform duration (in multiples of 4 ns).
        area: The area under the waveform.
    """
    def __init__(self, duration, area):
        """Initializes a Blackman waveform."""
        super().__init__(duration)
        if area <= 0:
            raise ValueError("Area under the waveform needs to be positive.")
        self._area = area

    @property
    def duration(self):
        """The duration of the pulse (in multiples of 4 ns)."""
        return self._duration

    @property
    def samples(self):
        """The value at each time step that describes the waveform.

        Returns:
            numpy.ndarray: A numpy array with a value for each time step.
        """
        samples = np.clip(np.blackman(self._duration), 0, np.inf)
        scaling = self._area / np.sum(samples) / 1e-3
        return samples * scaling

    @property
    def first_value(self):
        """The first value in the waveform."""
        return 0

    @property
    def last_value(self):
        """The last value in the waveform."""
        return 0

    def __str__(self):
        return f"Blackman(Area: {self._area:.3g})"

    def __repr__(self):
        return f"BlackmanWaveform({self._duration} ns, Area: {self._area:.3g})"



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import matplotlib.pyplot as plt
import numpy as np


class Register:
    """A quantum register containing a set of qubits.

    Args:
        qubits (dict): Dictionary with the qubit names as keys and their
            position coordinates (in μm) as values
            (e.g. {'q0':(2, -1, 0), 'q1':(-5, 10, 0), ...}).
    """

    def __init__(self, qubits):
        """Initializes a custom Register."""
        if not isinstance(qubits, dict):
            raise TypeError("The qubits have to be stored in a dictionary "
                            "matching qubit ids to position coordinates.")
        self._ids = list(qubits.keys())
        self._coords = list(qubits.values())

    @property
    def qubits(self):
        """Dictionary of the qubit names and their position coordinates."""
        return dict(zip(self._ids, self._coords))

    @classmethod
    def from_coordinates(cls, coords, center=True, prefix=None):
        """Creates the register from an array of coordinates.

        Args:
            coords (ndarray): The coordinates of each qubit to include in the
                register.

        Keyword args:
            center(defaut=True): Whether or not to center the entire array
                around the origin.
            prefix (str): The prefix for the qubit ids. If defined, each qubit
                id starts with the prefix, followed by an int from 0 to N-1
                (e.g. prefix='q' -> IDs: 'q0', 'q1', 'q2', ...).
        """
        if center:
            coords = coords - np.mean(coords, axis=0)      # Centers the array
        if prefix is not None:
            pre = str(prefix)
            qubits = {pre+str(i): pos for i, pos in enumerate(coords)}
        else:
            qubits = dict(enumerate(coords))
        return cls(qubits)

    @classmethod
    def rectangle(cls, rows, columns, spacing=4, prefix=None):
        """Initializes the register with the qubits in a rectangular array.

        Args:
            rows (int): Number of rows.
            columns (int): Number of columns.

        Keyword args:
            spacing(float): The distance between neighbouring qubits in μm.
            prefix (str): The prefix for the qubit ids. If defined, each qubit
                id starts with the prefix, followed by an int from 0 to N-1
                (e.g. prefix='q' -> IDs: 'q0', 'q1', 'q2', ...)
        """
        coords = np.array([(x, y) for y in range(rows)
                           for x in range(columns)], dtype=float) * spacing

        return cls.from_coordinates(coords, center=True, prefix=prefix)

    @classmethod
    def square(cls, side, spacing=4, prefix=None):
        """Initializes the register with the qubits in a square array.

        Args:
            side (int): Side of the square in number of qubits.

        Keyword args:
            spacing(float): The distance between neighbouring qubits in μm.
            prefix (str): The prefix for the qubit ids. If defined, each qubit
                id starts with the prefix, followed by an int from 0 to N-1
                (e.g. prefix='q' -> IDs: 'q0', 'q1', 'q2', ...).
        """
        return cls.rectangle(side, side, spacing=spacing, prefix=prefix)

    @classmethod
    def triangular_lattice(cls, rows, atoms_per_row, spacing=4, prefix=None):
        """Initializes the register with the qubits in a triangular lattice.

        Initializes the qubits in a triangular lattice pattern, more
        specifically a triangular lattice with horizontal rows, meaning the
        triangles are pointing up and down.

        Args:
            rows (int): Number of rows.
            atoms_per_row (int): Number of atoms per row.

        Keyword args:
            spacing(float): The distance between neighbouring qubits in μm.
            prefix (str): The prefix for the qubit ids. If defined, each qubit
                id starts with the prefix, followed by an int from 0 to N-1
                (e.g. prefix='q' -> IDs: 'q0', 'q1', 'q2', ...).
        """
        coords = np.array([(x, y) for y in range(rows)
                           for x in range(atoms_per_row)], dtype=float)
        coords[:, 0] += 0.5 * np.mod(coords[:, 1], 2)
        coords[:, 1] *= np.sqrt(3) / 2
        coords *= spacing

        return cls.from_coordinates(coords, center=True, prefix=prefix)

    def rotate(self, degrees):
        """Rotates the array around the origin by the given angle.

        Args:
            degrees (float): The angle of rotation in degrees.
        """
        theta = np.deg2rad(degrees)
        rot = np.array([[np.cos(theta), -np.sin(theta)],
                        [np.sin(theta), np.cos(theta)]])
        self._coords = [rot @ v for v in self._coords]

    def draw(self, with_labels=True):
        """Draws the entire register.

        Keyword args:
            with_labels(bool, default=True): If True, writes the qubit ID's
                next to each qubit.
        """
        pos = np.array(self._coords)
        diffs = np.max(pos, axis=0) - np.min(pos, axis=0)
        diffs[diffs < 9] *= 1.5
        diffs[diffs < 9] += 2
        big_side = max(diffs)
        proportions = diffs / big_side
        Ls = proportions * min(big_side/4, 10)  # Figsize is, at most, (10,10)

        fig, ax = plt.subplots(figsize=Ls)
        ax.scatter(pos[:, 0], pos[:, 1], s=30, alpha=0.7,
                   c='darkgreen')
        ax.axvline(0, c='grey', alpha=0.5, linestyle=':')
        ax.axhline(0, c='grey', alpha=0.5, linestyle=':')
        ax.set_xlabel("µm")
        ax.set_ylabel("µm")
        ax.axis('equal')
        ax.spines['right'].set_color('none')
        ax.spines['top'].set_color('none')

        if with_labels:
            for q, coords in zip(self._ids, self._coords):
                ax.annotate(q, coords, fontsize=12, ha='left', va='bottom')

        plt.show()



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import qutip
import numpy as np


class SimulationResults:
    """Results of a simulation run of a pulse sequence.

    Contains methods for studying the states and extracting useful information
    from them.
    """

    def __init__(self, run_output, dim, size, basis_name, meas_basis=None):
        """Initializes a new SimulationResults instance.

        Args:
            run_output (list of qutip.Qobj): List of `qutip.Qobj` corresponding
                to the states at each time step after the evolution has been
                simulated.
            dim (int): The dimension of the local space of each atom (2 or 3).
            size (int): The number of atoms in the register.
            basis_name (str): The basis indicating the addressed atoms after
                the pulse sequence ('ground-rydberg', 'digital' or 'all').

        Keyword Args:
            meas_basis (None or str): The basis in which a sampling measurement
                is desired.
        """
        self._states = run_output
        self._dim = dim
        self._size = size
        if basis_name not in {'ground-rydberg', 'digital', 'all'}:
            raise ValueError(
                "`basis_name` must be 'ground-rydberg', 'digital' or 'all'."
                )
        self._basis_name = basis_name
        if meas_basis:
            if meas_basis not in {'ground-rydberg', 'digital'}:
                raise ValueError(
                    "`meas_basis` must be 'ground-rydberg' or 'digital'."
                    )
        self._meas_basis = meas_basis

    @property
    def states(self):
        """List of ``qutip.Qobj`` for each state in the simulation."""
        return list(self._states)

    def expect(self, obs_list):
        """Calculates the expectation value of a list of observables.

        Args:
            obs_list (array-like of qutip.Qobj or array-like of numpy.ndarray):
                A list of observables whose expectation value will be
                calculated. If necessary, each member will be transformed into
                a qutip.Qobj instance.
        """
        if not isinstance(obs_list, (list, np.ndarray)):
            raise TypeError("`obs_list` must be a list of operators")

        qobj_list = []
        for obs in obs_list:
            if not (isinstance(obs, np.ndarray)
                    or isinstance(obs, qutip.Qobj)):
                raise TypeError("Incompatible type of observable.")
            if obs.shape != (self._dim**self._size, self._dim**self._size):
                raise ValueError("Incompatible shape of observable.")
            # Transfrom to qutip.Qobj and take dims from state
            dim_list = [self._states[0].dims[0], self._states[0].dims[0]]
            qobj_list.append(qutip.Qobj(obs, dims=dim_list))

        return [qutip.expect(qobj, self._states) for qobj in qobj_list]

    def sample_final_state(self, meas_basis=None, N_samples=1000):
        r"""Returns the result of multiple measurements in a given basis.

        The enconding of the results depends on the meaurement basis. Namely:

        - *ground-rydberg* : :math:`1 = |r\rangle;~ 0 = |g\rangle, |h\rangle`
        - *digital* : :math:`1 = |h\rangle;~ 0 = |g\rangle, |r\rangle`

        Note:
            The results are presented using a big-endian representation,
            according to the pre-established qubit ordering in the register.
            This means that, when sampling a register with qubits ('q0','q1',
            ...), in this order, the corresponding value, in binary, will be
            0Bb0b1..., where b0 is the outcome of measuring 'q0', 'b1' of
            measuring 'q1' and so on.

        Keyword Args:
            meas_basis (str, default=None): 'ground-rydberg' or 'digital'. If
                left as None, uses the measurement basis defined in the
                original sequence.
            N_samples (int, default=1000): Number of samples to take.

        Raises:
            ValueError: If trying to sample without a defined 'meas_basis' in
                the arguments when the original sequence is not measured.
        """
        if meas_basis is None:
            if self._meas_basis is None:
                raise ValueError(
                    "Can't accept an undefined measurement basis because the "
                    "original sequence has no measurement."
                    )
            meas_basis = self._meas_basis

        if meas_basis not in {'ground-rydberg', 'digital'}:
            raise ValueError(
                "'meas_basis' can only be 'ground-rydberg' or 'digital'."
                )

        N = self._size
        self.N_samples = N_samples
        probs = np.abs(self._states[-1])**2
        if self._dim == 2:
            if meas_basis == self._basis_name:
                # State vector ordered with r first for 'ground_rydberg'
                # e.g. N=2: [rr, rg, gr, gg] -> [11, 10, 01, 00]
                # Invert the order ->  [00, 01, 10, 11] correspondence
                weights = probs if meas_basis == 'digital' else probs[::-1]
            else:
                return {'0' * N: int(N_samples)}
            weights = weights.flatten()

        elif self._dim == 3:
            if meas_basis == 'ground-rydberg':
                one_state = 0       # 1 = |r>
                ex_one = slice(1, 3)
            elif meas_basis == 'digital':
                one_state = 2       # 1 = |h>
                ex_one = slice(0, 2)

            probs = probs.reshape([3]*N)
            weights = []
            for dec_val in range(2**N):
                ind = []
                for v in np.binary_repr(dec_val, width=N):
                    if v == '0':
                        ind.append(ex_one)
                    else:
                        ind.append(one_state)
                # Eg: 'digital' basis => |1> = index 2, |0> = index 0, 1 = 0:2
                # p_11010 = sum(probs[2, 2, 0:2, 2, 0:2])
                # We sum all probabilites that correspond to measuring 11010,
                # namely hhghg, hhrhg, hhghr, hhrhr
                weights.append(np.sum(probs[tuple(ind)]))
        else:
            raise NotImplementedError(
                "Cannot sample system with single-atom state vectors of "
                "dimension > 3."
                )
        dist = np.random.multinomial(N_samples, weights)
        return {np.binary_repr(i, N): dist[i] for i in np.nonzero(dist)[0]}



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple
from collections.abc import Iterable
import copy
import warnings

import numpy as np

import pulser
from pulser.pulse import Pulse
from pulser.devices import MockDevice
from pulser._seq_drawer import draw_sequence
from pulser.utils import validate_duration

# Auxiliary class to store the information in the schedule
_TimeSlot = namedtuple('_TimeSlot', ['type', 'ti', 'tf', 'targets'])


class Sequence:
    """A sequence of operations on a device.

    A sequence is composed by

        - The device in which we want to implement it
        - The register of qubits on which to act
        - The device's channels that are used
        - The schedule of operations on each channel

    Args:
        register(Register): The atom register on which to apply the pulses.
        device(PasqalDevice): A valid device in which to execute the Sequence
            (import it from ``pulser.devices``).
    """
    def __init__(self, register, device):
        """Initializes a new pulse sequence."""
        cond1 = device not in pulser.devices._valid_devices
        cond2 = device != MockDevice
        if cond1 and cond2:
            names = [d.name for d in pulser.devices._valid_devices]
            error_msg = ("The Sequence's device has to be imported from "
                         + "pasqal.devices. Choose 'MockDevice' or between the"
                         + " following real devices:\n" + "\n".join(names))
            raise ValueError(error_msg)
        # Checks if register is compatible with the device
        device.validate_register(register)

        self._register = register
        self._device = device
        self._channels = {}
        self._schedule = {}
        self._phase_ref = {}  # The phase reference of each channel
        # Stores the ids of selected channels and their declared names
        self._taken_channels = {}
        self._qids = set(self.qubit_info.keys())  # IDs of all qubits in device
        self._last_used = {}    # Last time each qubit was used, by basis
        self._last_target = {}  # Last time a target happened, by channel

    @property
    def qubit_info(self):
        """Dictionary with the qubit's IDs and positions."""
        return self._register.qubits

    @property
    def declared_channels(self):
        """Channels declared in this Sequence."""
        return dict(self._channels)

    @property
    def available_channels(self):
        """Channels still available for declaration."""
        return {id: ch for id, ch in self._device.channels.items()
                if id not in self._taken_channels
                or self._device == MockDevice}

    def current_phase_ref(self, qubit, basis='digital'):
        """Current phase reference of a specific qubit for a given basis.

        Args:
            qubit (str): The id of the qubit whose phase shift is desired.

        Keyword args:
            basis (str): The basis (i.e. electronic transition) the phase
                reference is associated with. Must correspond to the basis of a
                declared channel.

        Returns:
            float: Current phase reference of 'qubit' in 'basis'.

        """
        if qubit not in self._qids:
            raise ValueError("'qubit' must be the id of a qubit declared in "
                             "this sequence's device.")

        if basis not in self._phase_ref:
            raise ValueError("No declared channel targets the given 'basis'.")

        return self._phase_ref[basis][qubit].last_phase

    def declare_channel(self, name, channel_id, initial_target=None):
        """Declares a new channel to the Sequence.

        Args:
            name (str): Unique name for the channel in the sequence.
            channel_id (str): How the channel is identified in the device.
                Consult ``Sequence.available_channels`` to see which channel
                ID's are still available and the associated channel's
                description.

        Keyword Args:
            initial_target (set, default=None): For 'Local' adressing channels
                only. Declares the initial target of the channel. If left as
                None, the initial target will have to be set manually as the
                first addition to this channel.
        """

        if name in self._channels:
            raise ValueError("The given name is already in use.")

        if channel_id not in self._device.channels:
            raise ValueError("No channel %s in the device." % channel_id)

        if channel_id not in self.available_channels:
            raise ValueError("Channel %s is not available." % channel_id)

        ch = self._device.channels[channel_id]
        self._channels[name] = ch
        self._taken_channels[channel_id] = name
        self._schedule[name] = []
        self._last_target[name] = 0

        if ch.basis not in self._phase_ref:
            self._phase_ref[ch.basis] = {q: _PhaseTracker(0)
                                         for q in self._qids}
            self._last_used[ch.basis] = {q: 0 for q in self._qids}

        if ch.addressing == 'Global':
            self._add_to_schedule(name, _TimeSlot('target', -1, 0, self._qids))
        elif initial_target is not None:
            self.target(initial_target, name)

    def add(self, pulse, channel, protocol='min-delay'):
        """Adds a pulse to a channel.

        Args:
            pulse (pulser.Pulse): The pulse object to add to the channel.
            channel (str): The channel's name provided when declared.

        Keyword Args:
            protocol (default='min-delay'): Stipulates how to deal with
                eventual conflicts with other channels, specifically in terms
                of having to channels act on the same target simultaneously.

                - 'min-delay'
                    Before adding the pulse, introduces the smallest
                    possible delay that avoids all exisiting conflicts.
                - 'no-delay'
                    Adds the pulse to the channel, regardless of
                    existing conflicts.
                - 'wait-for-all'
                    Before adding the pulse, adds a delay that
                    idles the channel until the end of the other channels'
                    latest pulse.
        """

        last = self._last(channel)
        self._validate_pulse(pulse, channel)

        valid_protocols = ['min-delay', 'no-delay', 'wait-for-all']
        if protocol not in valid_protocols:
            raise ValueError(f"Invalid protocol '{protocol}', only accepts "
                             "protocols: " + ", ".join(valid_protocols))

        t0 = last.tf    # Preliminary ti
        basis = self._channels[channel].basis
        phase_barriers = [self._phase_ref[basis][q].last_time
                          for q in last.targets]
        current_max_t = max(t0, *phase_barriers)
        if protocol != 'no-delay':
            for ch, seq in self._schedule.items():
                if ch == channel:
                    continue
                for op in self._schedule[ch][::-1]:
                    if op.tf <= current_max_t:
                        break
                    if not isinstance(op.type, Pulse):
                        continue
                    if op.targets & last.targets or protocol == 'wait-for-all':
                        current_max_t = op.tf
                        break
        ti = current_max_t
        tf = ti + pulse.duration
        if ti > t0:
            self.delay(ti-t0, channel)

        prs = {self._phase_ref[basis][q].last_phase for q in last.targets}
        if len(prs) != 1:
            raise ValueError("Cannot do a multiple-target pulse on qubits "
                             "with different phase references for the same "
                             "basis.")
        else:
            phase_ref = prs.pop()

        if phase_ref != 0:
            # Has to copy to keep the original pulse intact
            pulse = copy.deepcopy(pulse)
            pulse.phase = (pulse.phase + phase_ref) % (2 * np.pi)

        self._add_to_schedule(channel, _TimeSlot(pulse, ti, tf, last.targets))

        for q in last.targets:
            if self._last_used[basis][q] < tf:
                self._last_used[basis][q] = tf

        if pulse.post_phase_shift:
            self.phase_shift(pulse.post_phase_shift, *last.targets,
                             basis=basis)

    def target(self, qubits, channel):
        """Changes the target qubit of a 'Local' channel.

        Args:
            qubits (hashable, iterable): The new target for this channel. Must
                correspond to a qubit ID in device or an iterable of qubit IDs,
                when multi-qubit adressing is possible.
            channel (str): The channel's name provided when declared. Must be
                a channel with 'Local' addressing.
         """

        if channel not in self._channels:
            raise ValueError("Use the name of a declared channel.")

        if isinstance(qubits, Iterable) and not isinstance(qubits, str):
            qs = set(qubits)
        else:
            qs = {qubits}

        if not qs.issubset(self._qids):
            raise ValueError("The given qubits have to belong to the device.")

        if self._channels[channel].addressing != 'Local':
            raise ValueError("Can only choose target of 'Local' channels.")
        elif len(qs) > self._channels[channel].max_targets:
            raise ValueError(
                "This channel can target at most "
                f"{self._channels[channel].max_targets} qubits at a time"
            )

        basis = self._channels[channel].basis
        phase_refs = {self._phase_ref[basis][q].last_phase for q in qs}
        if len(phase_refs) != 1:
            raise ValueError("Cannot target multiple qubits with different "
                             "phase references for the same basis.")

        try:
            last = self._last(channel)
            if last.targets == qs:
                warnings.warn("The provided qubits are already the target. "
                              "Skipping this target instruction.")
                return
            ti = last.tf
            retarget = self._channels[channel].retarget_time
            elapsed = ti - self._last_target[channel]
            delta = np.clip(retarget - elapsed, 0, retarget)
            if delta != 0:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    delta = validate_duration(np.clip(delta, 16, np.inf))
            tf = ti + delta

        except ValueError:
            ti = -1
            tf = 0

        self._last_target[channel] = tf
        self._add_to_schedule(channel, _TimeSlot('target', ti, tf, qs))

    def delay(self, duration, channel):
        """Idles a given channel for a specific duration.

        Args:
            duration (int): Time to delay (in multiples of 4 ns).
            channel (str): The channel's name provided when declared.
        """
        last = self._last(channel)
        ti = last.tf
        tf = ti + validate_duration(duration)
        self._add_to_schedule(channel,
                              _TimeSlot('delay', ti, tf, last.targets))

    def measure(self, basis='ground-rydberg'):
        """Measures in a valid basis.

        Args:
            basis (str): Valid basis for measurement (consult the
                'supported_bases' attribute of the selected device for
                the available options).
        """
        available = self._device.supported_bases
        if basis not in available:
            raise ValueError(f"The basis '{basis}' is not supported by the "
                             "selected device. The available options are: "
                             + ", ".join(list(available)))

        if hasattr(self, '_measurement'):
            raise SystemError("The sequence has already been measured.")

        self._measurement = basis

    def phase_shift(self, phi, *targets, basis='digital'):
        r"""Shifts the phase of a qubit's reference by 'phi', for a given basis.

        This is equivalent to an :math:`R_z(\phi)` gate (i.e. a rotation of the
        target qubit's state by an angle :math:`\phi` around the z-axis of the
        Bloch sphere).

        Args:
            phi (float): The intended phase shift (in rads).
            targets: The ids of the qubits on which to apply the phase
                shift.

        Keyword Args:
            basis(str): The basis (i.e. electronic transition) to associate
                the phase shift to. Must correspond to the basis of a declared
                channel.
        """
        if phi % (2*np.pi) == 0:
            warnings.warn("A phase shift of 0 is meaningless, "
                          "it will be ommited.")
            return
        if not set(targets) <= self._qids:
            raise ValueError("All given targets have to be qubit ids declared"
                             " in this sequence's device.")

        if basis not in self._phase_ref:
            raise ValueError("No declared channel targets the given 'basis'.")

        for q in targets:
            t = self._last_used[basis][q]
            new_phase = self._phase_ref[basis][q].last_phase + phi
            self._phase_ref[basis][q][t] = new_phase

    def align(self, *channels):
        """Aligns multiple channels in time.

        Introduces delays that align the provided channels with the one that
        finished the latest, such that the next action added to any of them
        will start right after the latest channel has finished.

        Args:
            channels (str): The names of the channels to align, as given upon
                declaration.
        """

        ch_set = set(channels)
        # channels have to be a subset of the declared channels
        if not ch_set <= set(self._channels):
            raise ValueError("All channel names must correspond to declared"
                             " channels.")
        if len(channels) != len(ch_set):
            raise ValueError("The same channel was provided more than once.")

        if len(channels) < 2:
            raise ValueError("Needs at least two channels for alignment.")

        last_ts = {id: self._last(id).tf for id in channels}
        tf = max(last_ts.values())

        for id in channels:
            delta = tf - last_ts[id]
            if delta > 0:
                self.delay(delta, id)

    def draw(self):
        """Draws the sequence in its current sequence."""
        draw_sequence(self)

    def __str__(self):
        full = ""
        pulse_line = "t: {}->{} | {} | Targets: {}\n"
        target_line = "t: {}->{} | Target: {} | Phase Reference: {}\n"
        delay_line = "t: {}->{} | Delay \n"
        # phase_line = "t: {} | Phase shift of: {:.3f} | Targets: {}\n"
        for ch, seq in self._schedule.items():
            basis = self._channels[ch].basis
            full += f"Channel: {ch}\n"
            first_slot = True
            for ts in seq:
                if ts.type == 'delay':
                    full += delay_line.format(ts.ti, ts.tf)
                    continue

                tgts = list(ts.targets)
                tgt_txt = ", ".join([str(t) for t in tgts])
                if isinstance(ts.type, Pulse):
                    full += pulse_line.format(ts.ti, ts.tf, ts.type, tgt_txt)
                elif ts.type == 'target':
                    phase = self._phase_ref[basis][tgts[0]][ts.tf]
                    if first_slot:
                        full += (f"t: 0 | Initial targets: {tgt_txt} | " +
                                 f"Phase Reference: {phase} \n")
                        first_slot = False
                    else:
                        full += target_line.format(ts.ti, ts.tf, tgt_txt,
                                                   phase)

            full += "\n"

        if hasattr(self, "_measurement"):
            full += f"Measured in basis: {self._measurement}"

        return full

    def _add_to_schedule(self, channel, timeslot):
        if hasattr(self, "_measurement"):
            raise SystemError("The sequence has already been measured. "
                              "Nothing more can be added.")
        self._schedule[channel].append(timeslot)

    def _last(self, channel):
        """Shortcut to last element in the channel's schedule."""
        if channel not in self._schedule:
            raise ValueError("Use the name of a declared channel.")
        try:
            return self._schedule[channel][-1]
        except IndexError:
            raise ValueError("The chosen channel has no target.")

    def _validate_pulse(self, pulse, channel):
        if not isinstance(pulse, Pulse):
            raise TypeError("pulse input must be of type Pulse, not of type "
                            "{}.".format(type(pulse)))

        ch = self._channels[channel]
        if np.any(pulse.amplitude.samples > ch.max_amp):
            raise ValueError("The pulse's amplitude goes over the maximum "
                             "value allowed for the chosen channel.")
        if np.any(np.round(np.abs(pulse.detuning.samples),
                           decimals=6) > ch.max_abs_detuning):
            raise ValueError("The pulse's detuning values go out of the range "
                             "allowed for the chosen channel.")


class _PhaseTracker:
    """Tracks a phase reference over time."""

    def __init__(self, initial_phase):
        self._times = [0]
        self._phases = [self._format(initial_phase)]

    @property
    def last_time(self):
        return self._times[-1]

    @property
    def last_phase(self):
        return self._phases[-1]

    def changes(self, ti, tf, time_scale=1):
        """Changes in phases within ]ti, tf]."""
        start, end = np.searchsorted(
                self._times, (ti * time_scale, tf * time_scale), side='right')
        for i in range(start, end):
            change = self._phases[i] - self._phases[i-1]
            yield (self._times[i] / time_scale, change)

    def _format(self, phi):
        return phi % (2 * np.pi)

    def __setitem__(self, t, phi):
        phase = self._format(phi)
        if t in self._times:
            ind = self._times.index(t)
            self._phases[ind] = phase
        else:
            ind = np.searchsorted(self._times, t, side='right')
            self._times.insert(ind, t)
            self._phases.insert(ind, phase)

    def __getitem__(self, t):
        ind = np.searchsorted(self._times, t, side='right') - 1
        return self._phases[ind]



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Tuple

import numpy as np
from scipy.spatial.distance import pdist

from pulser.channels import Channel
from pulser.register import Register


@dataclass(frozen=True, repr=False)
class PasqalDevice:
    r"""Definition of a Pasqal Device.

    Attributes:
        name: The name of the device.
        dimensions: Whether it supports 2D or 3D arrays.
        max_atom_num: Maximum number of atoms supported in an array.
        max_radial_distance: The furthest away an atom can be from the center
            of the array (in μm).
        min_atom_distance: The closest together two atoms can be (in μm).
        interaction_coeff: :math:`C_6/\hbar` (in :math:`\mu m^6 / \mu s`),
            which sets the van der Waals interaction strength between atoms in
            the Rydberg state.
    """

    name: str
    dimensions: int
    max_atom_num: int
    max_radial_distance: int
    min_atom_distance: int
    _channels: Tuple[Tuple[str, Channel]]
    interaction_coeff: float = 5008713.

    def __post_init__(self):
        # Hack to override the docstring of an instance
        self.__dict__["__doc__"] = self._specs(for_docs=True)

    @property
    def channels(self):
        """Dictionary of available channels on this device."""
        return dict(self._channels)

    @property
    def supported_bases(self):
        """Available electronic transitions for control and measurement."""
        return {ch.basis for ch in self.channels.values()}

    def print_specs(self):
        """Prints the device specifications."""
        title = f"{self.name} Specifications"
        header = ["-"*len(title), title, "-"*len(title)]
        print("\n".join(header))
        print(self._specs())

    def __repr__(self):
        return self.name

    def rydberg_blockade_radius(self, rabi_frequency):
        """Calculates the Rydberg blockade radius for a given Rabi frequency.

        Args:
            rabi_frequency(float): The Rabi frequency, in rad/µs.

        Returns:
            float: The rydberg blockade radius, in μm.
        """
        return (self.interaction_coeff/rabi_frequency)**(1/6)

    def validate_register(self, register):
        """Checks if 'register' is compatible with this device.

        Args:
            register(pulser.Register): The Register to validate.
        """
        if not isinstance(register, Register):
            raise TypeError("register has to be a pulser.Register instance.")

        atoms = list(register.qubits.values())
        if len(atoms) > self.max_atom_num:
            raise ValueError("Too many atoms in the array, the device accepts "
                             "at most {} atoms.".format(self.max_atom_num))
        for pos in atoms:
            if len(pos) != self.dimensions:
                raise ValueError("All qubit positions must be {}D "
                                 "vectors.".format(self.dimensions))

        if len(atoms) > 1:
            distances = pdist(atoms)  # Pairwise distance between atoms
            if np.min(distances) < self.min_atom_distance:
                raise ValueError("Qubit positions don't respect the minimal "
                                 "distance between atoms for this device.")

        if np.max(np.linalg.norm(atoms, axis=1)) > self.max_radial_distance:
            raise ValueError("All qubits must be at most {} μm away from the "
                             "center of the array.".format(
                                                    self.max_radial_distance))

    def _specs(self, for_docs=False):
        lines = [
            "\nRegister requirements:",
            f" - Dimensions: {self.dimensions}D",
            f" - Maximum number of atoms: {self.max_atom_num}",
            f" - Maximum distance from origin: {self.max_radial_distance} μm",
            (" - Minimum distance between neighbouring atoms: "
             + f"{self.min_atom_distance} μm"),
            "\nChannels:"
            ]

        ch_lines = []
        for name, ch in self._channels:
            if for_docs:
                ch_lines += [
                    f" - ID: '{name}'",
                    f"\t- Type: {ch.name} (*{ch.basis}* basis)",
                    f"\t- Addressing: {ch.addressing}",
                    ("\t" + r"- Maximum :math:`\Omega`:"
                     + f" {ch.max_amp:.4g} rad/µs"),
                    ("\t" + r"- Maximum :math:`|\delta|`:"
                     + f" {ch.max_abs_detuning:.4g} rad/µs")
                    ]
                if ch.addressing == "Local":
                    ch_lines += [
                        f"\t- Maximum time to retarget: {ch.retarget_time} ns",
                        f"\t- Maximum simultaneous targets: {ch.max_targets}"
                        ]
            else:
                ch_lines.append(f" - '{name}': {ch!r}")

        return "\n".join(lines + ch_lines)



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pulser.devices._pasqal_device import PasqalDevice
from pulser.channels import Rydberg, Raman


MockDevice = PasqalDevice(
            name="MockDevice",
            dimensions=2,
            max_atom_num=2000,
            max_radial_distance=1000,
            min_atom_distance=1,
            _channels=(
                ("rydberg_global", Rydberg.Global(1000, 200)),
                ("rydberg_local", Rydberg.Local(1000, 200, 0, 2000)),
                ("raman_global", Raman.Global(1000, 200)),
                ("raman_local", Raman.Local(1000, 200, 0, 2000)),
                )
            )



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import qutip
import numpy as np
from copy import deepcopy

from pulser import Pulse, Sequence
from pulser.simresults import SimulationResults


class Simulation:
    """Simulation of a pulse sequence using QuTiP.

    Creates a Hamiltonian object with the proper dimension according to the
    pulse sequence given, then provides a method to time-evolve an initial
    state using the QuTiP solvers.

    Args:
        sequence (Sequence): An instance of a Pulser Sequence that we
            want to simulate.

    Keyword Args:
        sampling_rate (float): The fraction of samples that we wish to
            extract from the pulse sequence to simulate. Has to be a
            value between 0.05 and 1.0
    """

    def __init__(self, sequence, sampling_rate=1.0):
        """Initialize the Simulation with a specific pulser.Sequence."""
        if not isinstance(sequence, Sequence):
            raise TypeError("The provided sequence has to be a valid "
                            "pulser.Sequence instance.")
        if not sequence._schedule:
            raise ValueError("The provided sequence has no declared channels.")
        if all(sequence._schedule[x][-1].tf == 0 for x in sequence._channels):
            raise ValueError("No instructions given for the channels in the "
                             "sequence.")
        self._seq = sequence
        self._qdict = self._seq.qubit_info
        self._size = len(self._qdict)
        self._tot_duration = max(
            [self._seq._last(ch).tf for ch in self._seq._schedule]
        )

        if not (0 < sampling_rate <= 1.0):
            raise ValueError("`sampling_rate` must be positive and "
                             "not larger than 1.0")
        if int(self._tot_duration*sampling_rate) < 4:
            raise ValueError("`sampling_rate` is too small, less than 4 data "
                             "points.")
        self.sampling_rate = sampling_rate

        self._qid_index = {qid: i for i, qid in enumerate(self._qdict)}
        self.samples = {addr: {basis: {}
                               for basis in ['ground-rydberg', 'digital']}
                        for addr in ['Global', 'Local']}
        self.operators = deepcopy(self.samples)

        self._extract_samples()
        self._build_basis_and_op_matrices()
        self._construct_hamiltonian()

    def _extract_samples(self):
        """Populate samples dictionary with every pulse in the sequence."""

        def prepare_dict():
            # Duration includes retargeting, delays, etc.
            return {'amp': np.zeros(self._tot_duration),
                    'det': np.zeros(self._tot_duration),
                    'phase': np.zeros(self._tot_duration)}

        def write_samples(slot, samples_dict):
            samples_dict['amp'][slot.ti:slot.tf] += slot.type.amplitude.samples
            samples_dict['det'][slot.ti:slot.tf] += slot.type.detuning.samples
            samples_dict['phase'][slot.ti:slot.tf] = slot.type.phase

        for channel in self._seq.declared_channels:
            addr = self._seq.declared_channels[channel].addressing
            basis = self._seq.declared_channels[channel].basis

            samples_dict = self.samples[addr][basis]

            if addr == 'Global':
                if not samples_dict:
                    samples_dict = prepare_dict()
                for slot in self._seq._schedule[channel]:
                    if isinstance(slot.type, Pulse):
                        write_samples(slot, samples_dict)

            elif addr == 'Local':
                for slot in self._seq._schedule[channel]:
                    if isinstance(slot.type, Pulse):
                        for qubit in slot.targets:  # Allow multiaddressing
                            if qubit not in samples_dict:
                                samples_dict[qubit] = prepare_dict()
                            write_samples(slot, samples_dict[qubit])

            self.samples[addr][basis] = samples_dict

    def _build_basis_and_op_matrices(self):
        """Determine dimension, basis and projector operators."""
        # No samples => Empty dict entry => False
        if (not self.samples['Global']['digital']
                and not self.samples['Local']['digital']):
            self.basis_name = 'ground-rydberg'
            self.dim = 2
            basis = ['r', 'g']
            projectors = ['gr', 'rr', 'gg']
        elif (not self.samples['Global']['ground-rydberg']
                and not self.samples['Local']['ground-rydberg']):
            self.basis_name = 'digital'
            self.dim = 2
            basis = ['g', 'h']
            projectors = ['hg', 'hh', 'gg']
        else:
            self.basis_name = 'all'  # All three states
            self.dim = 3
            basis = ['r', 'g', 'h']
            projectors = ['gr', 'hg', 'rr', 'gg', 'hh']

        self.basis = {b: qutip.basis(self.dim, i) for i, b in enumerate(basis)}
        self.op_matrix = {'I': qutip.qeye(self.dim)}

        for proj in projectors:
            self.op_matrix['sigma_' + proj] = (
                self.basis[proj[0]] * self.basis[proj[1]].dag()
            )

    def _build_operator(self, op_id, *qubit_ids, global_op=False):
        """Create qutip.Qobj with nontrivial action at *qubit_ids."""
        if global_op:
            return sum(self._build_operator(op_id, q_id)
                       for q_id in self._qdict)

        if len(set(qubit_ids)) < len(qubit_ids):
            raise ValueError("Duplicate atom ids in argument list.")
        # List of identity matrices with shape of operator:
        temp = [self.op_matrix['I'] for _ in range(self._size)]
        for q_id in qubit_ids:
            temp[self._qid_index[q_id]] = self.op_matrix[op_id]
        return qutip.tensor(temp)

    def _construct_hamiltonian(self):
        def adapt(full_array):
            """Adapt list to correspond to sampling rate"""
            indexes = np.linspace(0, self._tot_duration-1,
                                  int(self.sampling_rate*self._tot_duration),
                                  dtype=int)
            return full_array[indexes]

        def make_vdw_term():
            """Construct the Van der Waals interaction Term.

            For each pair of qubits, calculate the distance between them, then
            assign the local operator "sigma_rr" at each pair. The units are
            given so that the coefficient includes a 1/hbar factor.
            """
            vdw = 0
            # Get every pair without duplicates
            for q1, q2 in itertools.combinations(self._qdict.keys(), r=2):
                dist = np.linalg.norm(
                    self._qdict[q1] - self._qdict[q2])
                U = 0.5 * self._seq._device.interaction_coeff / dist**6
                vdw += U * self._build_operator('sigma_rr', q1, q2)
            return vdw

        def build_coeffs_ops(basis, addr):
            """Build coefficients and operators for the hamiltonian QobjEvo."""
            samples = self.samples[addr][basis]
            operators = self.operators[addr][basis]
            # Choose operator names according to addressing:
            if basis == 'ground-rydberg':
                op_ids = ['sigma_gr', 'sigma_rr']
            elif basis == 'digital':
                op_ids = ['sigma_hg', 'sigma_gg']

            terms = []
            if addr == 'Global':
                coeffs = [0.5*samples['amp'] * np.exp(-1j * samples['phase']),
                          -0.5 * samples['det']]
                for op_id, coeff in zip(op_ids, coeffs):
                    if np.any(coeff != 0):
                        # Build once global operators as they are needed
                        if op_id not in operators:
                            operators[op_id] =\
                                self._build_operator(op_id, global_op=True)
                        terms.append([operators[op_id], adapt(coeff)])
            elif addr == 'Local':
                for q_id, samples_q in samples.items():
                    if q_id not in operators:
                        operators[q_id] = {}
                    coeffs = [0.5*samples_q['amp'] *
                              np.exp(-1j * samples_q['phase']),
                              -0.5 * samples_q['det']]
                    for coeff, op_id in zip(coeffs, op_ids):
                        if np.any(coeff != 0):
                            if op_id not in operators[q_id]:
                                operators[q_id][op_id] = \
                                    self._build_operator(op_id, q_id)
                            terms.append([operators[q_id][op_id],
                                          adapt(coeff)])

            self.operators[addr][basis] = operators
            return terms

        # Time independent term:
        if self.basis_name == 'digital':
            qobj_list = []
        else:
            # Van der Waals Interaction Terms
            qobj_list = [make_vdw_term()]

        # Time dependent terms:
        for addr in self.samples:
            for basis in self.samples[addr]:
                if self.samples[addr][basis]:
                    qobj_list += build_coeffs_ops(basis, addr)

        self._times = adapt(np.arange(self._tot_duration,
                                      dtype=np.double)/1000)

        ham = qutip.QobjEvo(qobj_list, tlist=self._times)
        ham = ham + ham.dag()
        ham.compress()

        self._hamiltonian = ham

    # Run Simulation Evolution using Qutip
    def run(self, initial_state=None, progress_bar=None, **options):
        """Simulate the sequence using QuTiP's solvers.

        Keyword Args:
            initial_state (array): The initial quantum state of the
                           evolution. Will be transformed into a
                           qutip.Qobj instance.
            progress_bar (bool): If True, the progress bar of QuTiP's sesolve()
                        will be shown.

        Returns:
            SimulationResults: Object containing the time evolution results.
        """
        if initial_state is not None:
            if isinstance(initial_state, qutip.Qobj):
                if initial_state.shape != (self.dim**self._size, 1):
                    raise ValueError("Incompatible shape of initial_state")
                self._initial_state = initial_state
            else:
                if initial_state.shape != (self.dim**self._size,):
                    raise ValueError("Incompatible shape of initial_state")
                self._initial_state = qutip.Qobj(initial_state)
        else:
            # by default, initial state is "ground" state of g-r basis.
            all_ground = [self.basis['g'] for _ in range(self._size)]
            self._initial_state = qutip.tensor(all_ground)

        result = qutip.sesolve(self._hamiltonian,
                               self._initial_state,
                               self._times,
                               progress_bar=progress_bar,
                               options=qutip.Options(max_step=5,
                                                     **options)
                               )

        if hasattr(self._seq, '_measurement'):
            meas_basis = self._seq._measurement
        else:
            meas_basis = None

        return SimulationResults(
            result.states, self.dim, self._size, self.basis_name,
            meas_basis=meas_basis
        )



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Definitions of Pasqal's devices."""
import numpy as np

from pulser.devices._pasqal_device import PasqalDevice
from pulser.channels import Raman, Rydberg


Chadoq2 = PasqalDevice(
    name="Chadoq2",
    dimensions=2,
    max_atom_num=100,
    max_radial_distance=50,
    min_atom_distance=4,
    _channels=(
        ("rydberg_global", Rydberg.Global(2 * np.pi * 20, 2 * np.pi * 2.5)),
        ("rydberg_local", Rydberg.Local(2 * np.pi * 20, 2 * np.pi * 10)),
        ("raman_local", Raman.Local(2 * np.pi * 20, 2 * np.pi * 10)),
    ),
)



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings


def validate_duration(duration, min_duration=16, max_duration=67108864):
    """Validates a time interval.

    Returns:
        int: The duration in multiples of 4 ns.
    """
    try:
        _duration = int(duration)
    except (TypeError, ValueError):
        raise TypeError("duration needs to be castable to an int but "
                        "type %s was provided" % type(duration))

    if duration < min_duration:
        raise ValueError(f"duration has to be at least {min_duration} ns.")

    if duration > max_duration:
        raise ValueError(f"duration can be at most {max_duration} ns.")

    if duration % 4 != 0:
        _duration -= _duration % 4
        warnings.warn("The given duration is below the machine's precision"
                      " of 4 ns time steps. It was rounded down to the"
                      " nearest multiple of 4 ns.")
    return _duration



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import matplotlib.pyplot as plt
import numpy as np

from pulser.waveforms import Waveform, ConstantWaveform


class Pulse:
    r"""A generic pulse.

    In Pulser, a Pulse is a modulation of a frequency signal in amplitude
    and/or frequency, with a specific phase, over a given duration. Amplitude
    and frequency modulation are defined by :class:`Waveform` child classes.
    Frequency modulation is determined by a detuning waveform, which describes
    the shift in frequency from the channel's central frequency over time.
    If either quantity is constant throughout the entire pulse, use the
    ``ConstantDetuning``, ``ConstantAmplitude`` or ``ConstantPulse`` class
    method to create it.

    Note:
        We define the ``amplitude`` of a pulse to be its Rabi frequency,
        :math:`\Omega`, in rad/µs. Equivalently, the ``detuning`` is
        :math:`\delta`, also in rad/µs.

    Args:
        amplitude (Waveform): The pulse amplitude waveform.
        detuning (Waveform): The pulse detuning waveform.
        phase (float): The pulse phase (in radians).

    Keyword Args:
        post_phase_shift (default=0): Optionally lets you add a phase shift
            (in rads) immediately after the end of the pulse. This allows for
            enconding of arbitrary single-qubit gates into a single pulse
            (see ``Sequence.phase_shift()`` for more information).
    """

    def __init__(self, amplitude, detuning, phase, post_phase_shift=0):
        """Initializes a new Pulse."""

        if not (isinstance(amplitude, Waveform) and
                isinstance(detuning, Waveform)):
            raise TypeError("'amplitude' and 'detuning' have to be waveforms.")

        if detuning.duration != amplitude.duration:
            raise ValueError(
                "Detuning and amplitude waveforms' durations don't match.")
        self.duration = amplitude.duration
        if np.any(amplitude.samples < 0):
            raise ValueError("An amplitude waveform has always to be "
                             "non-negative.")
        self.amplitude = amplitude
        self.detuning = detuning
        self.phase = float(phase) % (2 * np.pi)
        self.post_phase_shift = float(post_phase_shift) % (2 * np.pi)

    @classmethod
    def ConstantDetuning(cls, amplitude, detuning, phase, post_phase_shift=0):
        """Pulse with an amplitude waveform and a constant detuning.

        Args:
            amplitude (Waveform): The pulse amplitude waveform.
            detuning (float): The detuning value (in rad/µs).
            phase (float): The pulse phase (in radians).
        """

        detuning_wf = ConstantWaveform(amplitude.duration, detuning)
        return cls(amplitude, detuning_wf, phase, post_phase_shift)

    @classmethod
    def ConstantAmplitude(cls, amplitude, detuning, phase, post_phase_shift=0):
        """Pulse with a constant amplitude and a detuning waveform.

        Args:
            amplitude (float): The pulse amplitude value (in rad/µs).
            detuning (Waveform): The pulse detuning waveform.
            phase (float): The pulse phase (in radians).
        """

        amplitude_wf = ConstantWaveform(detuning.duration, amplitude)
        return cls(amplitude_wf, detuning, phase, post_phase_shift)

    @classmethod
    def ConstantPulse(cls, duration, amplitude, detuning, phase,
                      post_phase_shift=0):
        """Pulse with a constant amplitude and a constant detuning.

        Args:
            duration (int): The pulse duration (in multiples of 4 ns).
            amplitude (float): The pulse amplitude value (in rad/µs).
            detuning (float): The detuning value (in rad/µs).
            phase (float): The pulse phase (in radians).
        """

        amplitude_wf = ConstantWaveform(duration, amplitude)
        detuning_wf = ConstantWaveform(duration, detuning)
        return cls(amplitude_wf, detuning_wf, phase, post_phase_shift)

    def draw(self):
        """Draws the pulse's amplitude and frequency waveforms."""

        fig, ax1 = plt.subplots()
        ax2 = ax1.twinx()

        self.amplitude._plot(ax1, r"$\Omega$ (rad/µs)", color="darkgreen")
        self.detuning._plot(ax2, r"$\delta$ (rad/µs)", color="indigo")

        fig.tight_layout()
        plt.show()

    def __str__(self):
        return "Pulse(Amp={!s}, Detuning={!s}, Phase={:.3g})".format(
            self.amplitude, self.detuning, self.phase)

    def __repr__(self):
        return (f"Pulse(amp={self.amplitude!r}, detuning={self.detuning!r}, " +
                f"phase={self.phase:.3g}, " +
                f"post_phase_shift={self.post_phase_shift:.3g})")



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import ClassVar


@dataclass(init=False, repr=False, frozen=True)
class Channel:
    """Base class of a hardware channel.

    Not to be initialized itself, but rather through a child class and the
    ``Local`` or ``Global`` classmethods.

    Example:
        To create a channel targeting the 'ground-rydberg' transition globally,
        call ``Rydberg.Global(...)``.
    """

    name: ClassVar[str]
    basis: ClassVar[str]
    addressing: str
    max_abs_detuning: float
    max_amp: float
    retarget_time: int = None
    max_targets: int = 1

    @classmethod
    def Local(cls, max_abs_detuning, max_amp, retarget_time=220,
              max_targets=1):
        """Initializes the channel with local addressing.

        Args:
            max_abs_detuning (float): Maximum possible detuning (in rad/µs), in
                absolute value.
            max_amp(float): Maximum pulse amplitude (in rad/µs).
            retarget_time (int): Maximum time to change the target (in ns).

        Keyword Args:
            max_targets (int, default=1): How many qubits can be addressed at
                once by the same beam.
        """

        return cls('Local', max_abs_detuning, max_amp, max_targets=max_targets,
                   retarget_time=retarget_time)

    @classmethod
    def Global(cls, max_abs_detuning, max_amp):
        """Initializes the channel with global addressing.

        Args:
            max_abs_detuning (tuple): Maximum possible detuning (in rad/µs), in
                absolute value.
            max_amp(tuple): Maximum pulse amplitude (in rad/µs).
        """

        return cls('Global', max_abs_detuning, max_amp)

    def __repr__(self):
        s = ".{}(Max Absolute Detuning: {} rad/µs, Max Amplitude: {} rad/µs"
        config = s.format(self.addressing, self.max_abs_detuning, self.max_amp)
        if self.addressing == 'Local':
            config += f", Target time: {self.retarget_time} ns"
            if self.max_targets > 1:
                config += f", Max targets: {self.max_targets}"
        config += f", Basis: '{self.basis}'"
        return self.name + config + ")"


@dataclass(init=True, repr=False, frozen=True)
class Raman(Channel):
    """Raman beam channel.

    Channel targeting the transition between the hyperfine ground states, in
    which the 'digital' basis is encoded. See base class.
    """
    name: ClassVar[str] = 'Raman'
    basis: ClassVar[str] = 'digital'


@dataclass(init=True, repr=False, frozen=True)
class Rydberg(Channel):
    """Rydberg beam channel.

    Channel targeting the transition between the ground and rydberg states,
    thus enconding the 'ground-rydberg' basis. See base class.
    """
    name: ClassVar[str] = 'Rydberg'
    basis: ClassVar[str] = 'ground-rydberg'



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages

setup(
    name="pulser",
    version="0.1.0",
    install_requires=[
        "matplotlib",
        "numpy",
        "scipy",
        "qutip",
    ],
    packages=find_packages(),
    include_package_data=True,
    description="A pulse-level composer for neutral-atom quantum devices.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Pulser Development Team",
    python_requires=">=3.7.0",
    license="Apache 2.0",
    classifiers=[
         "Development Status :: 3 - Alpha",
         "Programming Language :: Python :: 3",
         ],
    url="https://github.com/pasqal-io/Pulser",
)



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import matplotlib.pyplot as plt
import numpy as np

from pulser.waveforms import ConstantWaveform


def gather_data(seq):
    """Collects the whole sequence data for plotting.

    Args:
        seq (pulser.Sequence): The input sequence of operations on a device.

    Returns:
        dict: The data to plot.
    """
    # The minimum time axis length is 100 ns
    seq._total_duration = max([seq._last(ch).tf for ch in seq._schedule
                               if seq._schedule[ch]] + [100])
    data = {}
    for ch, sch in seq._schedule.items():
        time = [-1]     # To not break the "time[-1]" later on
        amp = []
        detuning = []
        target = {}
        # phase_shift = {}
        for slot in sch:
            if slot.ti == -1:
                target['initial'] = slot.targets
                time += [0]
                amp += [0]
                detuning += [0]
                continue
            if slot.type in ['delay', 'target']:
                time += [slot.ti, slot.tf-1]
                amp += [0, 0]
                detuning += [0, 0]
                if slot.type == 'target':
                    target[(slot.ti, slot.tf-1)] = slot.targets
                continue
            pulse = slot.type
            if (isinstance(pulse.amplitude, ConstantWaveform) and
                    isinstance(pulse.detuning, ConstantWaveform)):
                time += [slot.ti, slot.tf-1]
                amp += [pulse.amplitude._value] * 2
                detuning += [pulse.detuning._value] * 2
            else:
                time += list(range(slot.ti, slot.tf))
                amp += pulse.amplitude.samples.tolist()
                detuning += pulse.detuning.samples.tolist()
        if time[-1] < seq._total_duration - 1:
            time += [time[-1]+1, seq._total_duration-1]
            amp += [0, 0]
            detuning += [0, 0]
        # Store everything
        time.pop(0)     # Removes the -1 in the beginning
        data[ch] = {'time': time, 'amp': amp, 'detuning': detuning,
                    'target': target}
        if hasattr(seq, "_measurement"):
            data[ch]['measurement'] = seq._measurement
    return data


def draw_sequence(seq):
    """Draw the entire sequence.

    Args:
        seq (pulser.Sequence): The input sequence of operations on a device.
    """

    def phase_str(phi):
        """Formats a phase value for printing."""
        value = (((phi + np.pi) % (2*np.pi)) - np.pi) / np.pi
        if value == -1:
            return r"$\pi$"
        elif value == 0:
            return "0"
        else:
            return r"{:.2g}$\pi$".format(value)

    n_channels = len(seq._channels)
    if not n_channels:
        raise SystemError("Can't draw an empty sequence.")
    data = gather_data(seq)
    time_scale = 1e3 if seq._total_duration > 1e4 else 1

    # Boxes for qubit and phase text
    q_box = dict(boxstyle="round", facecolor='orange')
    ph_box = dict(boxstyle="round", facecolor='ghostwhite')

    fig = plt.figure(constrained_layout=False, figsize=(20, 4.5*n_channels))
    gs = fig.add_gridspec(n_channels, 1, hspace=0.075)

    ch_axes = {}
    for i, (ch, gs_) in enumerate(zip(seq._channels, gs)):
        ax = fig.add_subplot(gs_)
        ax.spines['top'].set_color('none')
        ax.spines['bottom'].set_color('none')
        ax.spines['left'].set_color('none')
        ax.spines['right'].set_color('none')
        ax.tick_params(labelcolor='w', top=False, bottom=False, left=False,
                       right=False)
        ax.set_ylabel(ch, labelpad=40, fontsize=18)
        subgs = gs_.subgridspec(2, 1, hspace=0.)
        ax1 = fig.add_subplot(subgs[0, :])
        ax2 = fig.add_subplot(subgs[1, :])
        ch_axes[ch] = (ax1, ax2)
        for j, ax in enumerate(ch_axes[ch]):
            ax.axvline(0, linestyle='--', linewidth=0.5, color='grey')
            if j == 0:
                ax.spines['bottom'].set_visible(False)
            else:
                ax.spines['top'].set_visible(False)

            if i < n_channels - 1 or j == 0:
                ax.tick_params(axis='x', which='both', bottom=True,
                               top=False, labelbottom=False, direction='in')
            else:
                unit = 'ns' if time_scale == 1 else r'$\mu s$'
                ax.set_xlabel(f't ({unit})', fontsize=12)

    for ch, (a, b) in ch_axes.items():
        basis = seq._channels[ch].basis
        t = np.array(data[ch]['time']) / time_scale
        ya = data[ch]['amp']
        yb = data[ch]['detuning']

        t_min = -t[-1]*0.03
        t_max = t[-1]*1.05
        a.set_xlim(t_min, t_max)
        b.set_xlim(t_min, t_max)

        max_amp = np.max(ya)
        max_amp = 1 if max_amp == 0 else max_amp
        amp_top = max_amp * 1.2
        a.set_ylim(-0.02, amp_top)
        det_max = np.max(yb)
        det_min = np.min(yb)
        det_range = det_max - det_min
        if det_range == 0:
            det_min, det_max, det_range = -1, 1, 2
        det_top = det_max + det_range * 0.15
        det_bottom = det_min - det_range * 0.05
        b.set_ylim(det_bottom, det_top)

        a.plot(t, ya, color="darkgreen", linewidth=0.8)
        b.plot(t, yb, color='indigo', linewidth=0.8)
        a.fill_between(t, 0, ya, color="darkgreen", alpha=0.3)
        b.fill_between(t, 0, yb, color="indigo", alpha=0.3)
        a.set_ylabel(r'$\Omega$ (rad/µs)', fontsize=14, labelpad=10)
        b.set_ylabel(r'$\delta$ (rad/µs)', fontsize=14)

        target_regions = []     # [[start1, [targets1], end1],...]
        for coords in data[ch]['target']:
            targets = list(data[ch]['target'][coords])
            tgt_strs = [str(q) for q in targets]
            tgt_txt_y = max_amp*1.1-0.25*(len(targets)-1)
            tgt_str = "\n".join(tgt_strs)
            if coords == 'initial':
                x = t_min + t[-1]*0.005
                target_regions.append([0, targets])
                if seq._channels[ch].addressing == 'Global':
                    a.text(x, amp_top*0.98, "GLOBAL", fontsize=13,
                           rotation=90, ha='left', va='top', bbox=q_box)
                else:
                    a.text(x, tgt_txt_y, tgt_str, fontsize=12, ha='left',
                           bbox=q_box)
                    phase = seq._phase_ref[basis][targets[0]][0]
                    if phase:
                        msg = r"$\phi=$" + phase_str(phase)
                        a.text(0, max_amp*1.1, msg, ha='left', fontsize=12,
                               bbox=ph_box)
            else:
                ti, tf = np.array(coords) / time_scale
                target_regions[-1].append(ti)   # Closing previous regions
                target_regions.append([tf + 1/time_scale, targets])  # New one
                phase = seq._phase_ref[basis][targets[0]][tf * time_scale + 1]
                a.axvspan(ti, tf, alpha=0.4, color='grey', hatch='//')
                b.axvspan(ti, tf, alpha=0.4, color='grey', hatch='//')
                a.text(tf + t[-1]*5e-3, tgt_txt_y, tgt_str, ha='left',
                       fontsize=12, bbox=q_box)
                if phase:
                    msg = r"$\phi=$" + phase_str(phase)
                    wrd_len = len(max(tgt_strs, key=len))
                    x = tf + t[-1]*0.01*(wrd_len+1)
                    a.text(x, max_amp*1.1, msg, ha='left',
                           fontsize=12, bbox=ph_box)
        # Terminate the last open regions
        if target_regions:
            target_regions[-1].append(t[-1])
        for start, targets, end in target_regions:
            q = targets[0]  # All targets have the same ref, so we pick
            ref = seq._phase_ref[basis][q]
            if end != seq._total_duration - 1 or 'measurement' not in data[ch]:
                end += 1 / time_scale
            for t_, delta in ref.changes(start, end, time_scale=time_scale):
                conf = dict(linestyle='--', linewidth=1.5, color='black')
                a.axvline(t_, **conf)
                b.axvline(t_, **conf)
                msg = u"\u27F2 " + phase_str(delta)
                a.text(t_-t[-1]*8e-3, max_amp*1.1, msg, ha='right',
                       fontsize=14, bbox=ph_box)

        if 'measurement' in data[ch]:
            msg = f"Basis: {data[ch]['measurement']}"
            b.text(t[-1]*1.025, det_top, msg, ha='center', va='center',
                   fontsize=14, color='white', rotation=90)
            a.axvspan(t[-1], t_max, color='midnightblue', alpha=1)
            b.axvspan(t[-1], t_max, color='midnightblue', alpha=1)
            a.axhline(0, xmax=0.95, linestyle='-', linewidth=0.5,
                      color='grey')
            b.axhline(0, xmax=0.95, linestyle=':', linewidth=0.5,
                      color='grey')
        else:
            a.axhline(0, linestyle='-', linewidth=0.5, color='grey')
            b.axhline(0, linestyle=':', linewidth=0.5, color='grey')

    plt.show()
